r"""
Sensitivity curves
==================

This package provides functions to compute sensitivity curves from noise and
signal covariance matrices.

We generalize the definition of the sensitivity, sometimes called the combined
or optimal sensitivity, for a set of (potentially noise and signal) observables.
The sensitivity is defined as the trace of the product of the signal covariance
matrix (normalized by the strain power used to compute it, i.e., the trace of
the strain covariance matrix) and the inverse of the noise covariance matrix,

.. math::

    S(t, f) = \frac{1}{\text{Tr}[\mathbf{C}_\text{strain}]}
    \text{Tr}[\mathbf{C}_\text{signal}(t, f) \mathbf{C}_\text{noise}^{-1}(t,
    f)].

Note that the normalized signal covariance matrix is often interpreted as the
response matrix.

To compute single-channel sensitivity curves, one can trim the covariance
matrices to the single channel of interest. For example, if the signal
covariance matrix is of shape ``(..., N, N)`` and the noise covariance matrix
is of shape ``(..., M, M)``, one can compute the sensitivity for the first
channel of the signal covariance matrix and the first channel of the noise
covariance matrix as follows:

.. code-block:: python

    import numpy as np
    from segwo.sensitivity import compute_sensitivity_from_covariances

    sensitivity = compute_sensitivity_from_covariances(
        noise_cov[..., 0:1, 0:1], signal_cov[..., 0:1, 0:1], strain_power
    )

Note that the sum of single-channel sensitivities is not equal to the combined
sensitivity in general.

Computing sensitivity
---------------------

.. autofunction:: compute_sensitivity_from_covariances

"""

import numpy as np


def compute_sensitivity_from_covariances(
    noise_cov: np.ndarray, signal_cov: np.ndarray, strain_power: float = 1
) -> np.ndarray:
    r"""Compute sensitivity curve from noise and signal covariance matrices.

    The sensitivity is define as the trace of the product of the signal
    covariance matrix (normalized by the strain power used to compute it, i.e.,
    the trace of the strain covariance matrix) and the inverse of the noise
    covariance matrix,

    .. math::

        S(t, f) = \frac{1}{\text{Tr}[\mathbf{C}_\text{strain}]}
        \text{Tr}[\mathbf{C}_\text{signal}(t, f) \mathbf{C}_\text{noise}^{-1}(t,
        f)].

    Note that the normalized signal covariance matrix is often interpreted as
    the response matrix, and the sensitivity is defined as the trace of the
    product of the response and the inverse of the noise covariance matrix.

    This definition is valid for any set of :math:`N` observables for which the
    product :math:`\mathbf{C}_\text{signal} \mathbf{C}_\text{noise}^{-1}` is
    full rank.

    Args:
        noise_cov: Noise covariance matrix, of shape ``(..., N, N)``.
        signal_cov: Signal covariance matrix, of shape ``(..., N, N)``.
        strain_power: Strain power used to compute the signal covariance matrix,
            i.e., the trace of the strain covariance matrix. This is used to
            normalized the signal covariance matrix.

    Returns:
        Sensitivity curve, as an array of shape ``(...)``. Those axes are
        usually reserved for time and frequency.

        Units are those of ``response_cov`` divided by those of ``noise_cov``
        and power spectral density [1/Hz].

    Raises:
        ValueError: If the noise covariance matrix is singular.
    """
    if noise_cov.ndim < 2:
        raise ValueError("Noise covariance must have at least two dimensions.")
    if noise_cov.shape[-1] != noise_cov.shape[-2]:
        raise ValueError("Noise covariance must be square.")
    if signal_cov.ndim < 2:
        raise ValueError("Response covariance must have at least two dimensions.")
    if signal_cov.shape[-1] != signal_cov.shape[-2]:
        raise ValueError("Response covariance must be square.")

    if noise_cov.shape[1] == 1:
        inv_noise_cov = 1 / noise_cov
    else:
        try:
            inv_noise_cov = np.linalg.inv(noise_cov)  # (..., N, N)
        except np.linalg.LinAlgError as e:
            raise ValueError("Noise covariance is singular.") from e

    snr_psd = np.einsum("...ij, ...ji -> ...", signal_cov, inv_noise_cov)  # (...)
    np.allclose(snr_psd.imag / np.abs(snr_psd), 0)  # imaginary part should be small

    normalized_snr_psd = snr_psd.real / strain_power
    return 1 / normalized_snr_psd



"""
Response functions
==================

This module contains functions to compute the response to gravitational waves
of a detector with an arbitrary number of interferometric links. The response is
formalised as a mixing matrix that relates the gravitational wave strain to the
fractional frequency deviation (aka, Doppler shift) of the links.

.. note::

    The response for an arbitrary set of observables is not yet implemented.
    Only the response for the LISA case (a triangle of 6 links) is available.

Strain and polarization
-----------------------

We define here the conventions for the polarization basis, as well as
transformation matrices between the linear and circular polarization bases.

.. autodata:: LINEAR_POLARIZATION

.. autodata:: CIRCULAR_POLARIZATION

.. autodata:: LINEAR2CIRCULAR

.. autodata:: CIRCULAR2LINEAR

Link response function
----------------------

.. autofunction:: compute_strain2link

.. autofunction:: compute_source_basis

Sky-averaged response function
------------------------------

.. autofunction:: compute_isotropic_signal_cov

.. autofunction:: sky_average

"""

from typing import Callable

import healpy as hp
import numpy as np
from lisaconstants import c

from segwo.cov import project_covariance

LINEAR_POLARIZATION = ["+", "x"]
"""Convention for linear (plus and cross) polarization basis."""

CIRCULAR_POLARIZATION = ["left", "right"]
"""Convention for circular (left and right) polarization basis."""

LINEAR2CIRCULAR = np.array([[1, -1j], [1, 1j]]) / np.sqrt(2)
"""Mixing matrix going from linear to circular polarization.

Order for rows and columns are given by :attr:`LINEAR_POLARIZATION` and
:attr:`CIRCULAR_POLARIZATION`, respectively.

Use :func:`segwo.cov.project_covariance` with this mixing matrix to
convert a linear polarization covariance matrix to a circular polarization
covariance matrix.
"""

CIRCULAR2LINEAR = np.array([[1, 1], [1j, -1j]]) / np.sqrt(2)
"""Mixing matrix going from circular to linear polarization.

Order for rows and columns are given by :attr:`CIRCULAR_POLARIZATION` and
:attr:`LINEAR_POLARIZATION`, respectively.

Use :func:`segwo.cov.project_covariance` with this mixing matrix to
convert a circular polarization covariance matrix to a linear polarization
covariance matrix.
"""


def compute_source_basis(
    beta: np.ndarray, lamb: np.ndarray
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Compute the source-localization basis vectors.

    Args:
        beta: Ecliptic latitude(s) [rad].
        lamb: Ecliptic longitude(s) [rad].

    Returns:
        A 3-tuple of vector basis ``(k, u, v)`` as defined by the LDC Manual
        (2018), of shape ``(len(beta), 3)``.

        First axis correspond to the common shape of the input arrays. Second
        axis is for the coordinates ``(x, y, z)``.

    Raises:
        ValueError: If ``beta`` and ``lamb`` are not scalars or 1D arrays of the
            same shape.
    """
    # Check shapes
    beta = np.atleast_1d(beta)
    lamb = np.atleast_1d(lamb)
    if beta.ndim > 1 or lamb.ndim > 1:
        raise ValueError("beta and lamb must be scalars or 1D arrays.")
    if beta.shape != lamb.shape:
        raise ValueError("beta and lamb must have the same shape.")

    k = np.array(
        [
            -np.cos(beta) * np.cos(lamb),
            -np.cos(beta) * np.sin(lamb),
            -np.sin(beta) * np.ones_like(lamb),
        ]
    )  # (xyz, loc)
    u = np.array(
        [
            np.ones_like(beta) * np.sin(lamb),
            np.ones_like(beta) * -np.cos(lamb),
            np.zeros_like(beta) * np.zeros_like(lamb),
        ]
    )  # (xyz, loc)
    v = np.array(
        [
            -np.sin(beta) * np.cos(lamb),
            -np.sin(beta) * np.sin(lamb),
            np.cos(beta) * np.ones_like(lamb),
        ]
    )  # (xyz, loc)

    return k.T, u.T, v.T  # (loc, xyz)


def compute_strain2link(
    f: np.ndarray,
    beta: np.ndarray,
    lamb: np.ndarray,
    ltts: np.ndarray,
    positions: np.ndarray,
    *,
    method: str = "baghi+23",
) -> np.ndarray:
    r"""Frequency-domain link response to a GW strain (mixing matrix).

    The response is computed for the linear (plus and cross) polarization basis.
    To use another polarization basis, multiply by the appropriate polarization
    basis transformation matrix.

    Here, ``beta`` and ``lamb`` should be scalars or arrays of the same shape.

    The response for the link :math:`ij` and for a polarization :math:`p` (plus
    or cross), is given by :cite:`Baghi:2023qnq` or :cite:`Hartwig:2023pft`. Use
    the ``method`` argument to select the method to use.

    .. rubric:: Method "baghi+23"

    From :cite:`Baghi:2023qnq`, the single link response is given in the time
    domain by

    .. math::

        y_{ij,p}(t, \vu{k}) = \frac{\xi_p(\vu{k})}{2(1 - \vu{n}_{ij}(t) \vdot
        \vu{k})} \qty[h_p\qty(t - \frac{L_{ij}(t)}{c} - \frac{\vu{k} \vdot
        \vb{x}_j(t)}{c}) - h_p\qty(t - \frac{\vu{k} \vdot \vb{x}_i(t)}{c})],

    where :math:`\vu{k}` is the unit vector pointing towards the source,
    :math:`\vu{n}_{ij}(t)` is the unit vector pointing from spacecraft :math:`j`
    to spacecraft :math:`i`, :math:`\vb{x}_i(t)` is the position of spacecraft
    :math:`i`, :math:`L_{ij}(t)` is the arm length, :math:`h_p` is the strain in
    polarization :math:`p` (plus or cross), at the Solar System Barycenter, and
    :math:`\xi_p(\vu{k})` is the antenna pattern function for polarization
    :math:`p`, given by

    .. math::

        \xi_+ = \qty(\vu{k} \vdot \vu{u})^2 - \qty(\vu{k} \vdot \vu{v})^2,
        \xi_\times = 2 \qty(\vu{k} \vdot \vu{u}) \qty(\vu{k} \vdot \vu{v}).

    Assuming a adiabatically static constellation, the response in the frequency
    domain is given by

    .. math::

        y_{ij,p}(t, f, \vu{k}) = h_p(t, f, \vu{k}) G_{ij,p}(t, f, \vu{k}),

    where :math:`G_{ij,p}(t, f, \vu{k})` is the single-link response kernel,
    which is what this function computes. It is given by

    .. math::

        G_{ij,p}(t, f, \vu{k}) = \frac{\xi_p(\vu{k})}{2(1 - \vu{n}_{ij}(t) \vdot
        \vu{k})} \qty[e^{-i 2\pi f \frac{L_{ij}(t) + \vu{k} \vdot
        \vb{x}_j(t)}{c}} - e^{-i 2\pi f \frac{\vu{k} \vdot \vb{x}_i(t)}{c}}].

    Note that we use here an equivalent factorized form of this kernel, in which
    a sinc function appears, in order to prevent any numerical issues when the
    denominator is close to zero.

    .. rubric:: Method "Hartwig+23"

    Taken from :cite:`Hartwig:2023pft`.

    The final equations 2.15, 2.16 are the sky average of the complex response
    covariance matrix. We compute here only the complex response, and take the
    scalar product and sky average in a second step.

    .. warning::

        There is a sign difference between the two methods, which has not been
        tracked down and corrected for. We therefore multiply the Hartwig+23
        response by -1 to match the Baghi+23 response.

    Args:
        f: Frequency grid [Hz].
        beta: Ecliptic latitude(s) [rad].
        lamb: Ecliptic longitude(s) [rad].
        ltts: Light travel times between the test masses [s], of shape
            ``(t, link)``.
        positions: Test-mass positions [m], of shape ``(t, tm, xyz)``.
        method: Computation method, either "baghi+23" or "hartwig+23".

    Returns:
        Mixing matrix, of shape ``(len(t), len(f), len(beta), 6, 2)``.

        First axis is for time. Second axis is for frequency. The third axis is
        for the source localization, in the order given by the input ``beta``
        and ``lamb``.

        Fourth axis corresponds to the links, in the order 12, 23, 31, 13, 32.
        Fifth axis is for the two polarizations, in the order given by
        :attr:`LINEAR_POLARIZATION`.

        The response is given in (dimensionless) fractional frequency
        deviations (i.e., :math:`\Delta \nu / \nu`).
    """
    if method == "baghi+23":
        return _link_response_baghi23(f, beta, lamb, ltts, positions)
    if method == "hartwig+23":
        return -1 * _link_response_hartwig23(f, beta, lamb, ltts, positions)
    raise ValueError("Method must be either 'baghi+23' or 'Hartwig+23'.")


def _link_response_baghi23(
    f: np.ndarray,
    beta: np.ndarray,
    lamb: np.ndarray,
    ltts: np.ndarray,
    positions: np.ndarray,
) -> np.ndarray:
    """Sky-dependent, frequency-domain link response.

    .. seealso:: :func:`link_response`
    """
    k, u, v = compute_source_basis(beta, lamb)  # (loc, xyz)

    # 6 links ordered as (12, 23, 31, 13, 32, 21)
    # (usual conventions, ij goes from j to i)
    index_emitter = np.array([1, 2, 0, 2, 1, 0])  # (link,)
    index_receiver = np.array([0, 1, 2, 0, 2, 1])  # (link,)
    n = positions[:, index_receiver] - positions[:, index_emitter]  # (t, link, xyz)
    n /= np.linalg.norm(n, axis=-1, keepdims=True)  # (t, link, xyz)

    # Dot products
    delay_link = np.einsum("tij, kj -> tki", n, k)  # (t, loc, link)
    delay_receiver = np.einsum(
        "tij, kj -> tki", positions[:, index_receiver], k
    )  # (t, loc, link)

    # Broadcast delay_receiver over frequencies
    delay_link = delay_link[:, None, :, :]  # (t, f, loc, link)
    delay_receiver = delay_receiver[:, None, :, :]  # (t, f, loc, link)

    # Broadcast f over locations and links
    f = f[:, None, None]  # (f, loc, link)

    # Broadcast ltts over frequencies and locations
    ltts = ltts[:, None, None, :]  # (t, f, loc, link)

    # Compute complex factor
    sinc_arg = f * ltts * (1 - delay_link)  # (t, f, loc, link)
    g = (
        (-1j * np.pi * f * ltts)  # (t, f, loc, link)
        * np.exp(-2j * np.pi * f * delay_receiver / c)  # (t, f, loc, link)
        * np.exp(-1j * np.pi * sinc_arg)  # (t, f, loc, link)
        * np.sinc(sinc_arg)  # (t, f, loc, link)
    )

    # Antenna pattern functions
    xi_plus = (
        np.einsum("tij, kj -> tki", n, u) ** 2 - np.einsum("tij, kj -> tki", n, v) ** 2
    )  # (t, loc, link)
    xi_cross = (
        2 * np.einsum("tij, kj -> tki", n, u) * np.einsum("tij, kj -> tki", n, v)
    )  # (t, loc, link)
    xi = np.stack([xi_plus, xi_cross], axis=-1)  # (t, loc, link, pol)

    return g[..., None] * xi[:, None]  # (t, f, loc, link, pol)


def _link_response_hartwig23(
    f: np.ndarray,
    beta: np.ndarray,
    lamb: np.ndarray,
    ltts: np.ndarray,
    positions: np.ndarray,
) -> np.ndarray:
    """Sky-dependent, frequency-domain link response.

    .. seealso:: :func:`link_response`
    """
    # Define coordinate system
    k, u, v = compute_source_basis(beta, lamb)  # (loc, xyz)
    # Define polarization tensors, eq. 2.10
    e_plus = np.einsum("li, lj -> lij", u, u) - np.einsum(
        "li, lj -> lij", v, v
    )  # (loc, xyz, xyz)
    e_cross = np.einsum("li, lj -> lij", u, v) + np.einsum(
        "li, lj -> lij", v, u
    )  # (loc, xyz, xyz)

    # Spacecraft positions, expressed in seconds
    positions /= c  # (t, sc, xyz)

    # 6 links ordered as (12, 23, 31, 13, 32, 21)
    # (usual conventions, ij goes from j to i)
    index_emitter = np.array([1, 2, 0, 2, 1, 0])  # (link,)
    index_receiver = np.array([0, 1, 2, 0, 2, 1])  # (link,)
    # Define link vectors, below eq. 2.11
    l_ij = positions[:, index_emitter] - positions[:, index_receiver]  # (t, link, xyz)
    n = l_ij / np.linalg.norm(l_ij, axis=-1, keepdims=True)  # (t, link, xyz)

    # Geometric factor, eq. 2.14
    g_plus = np.einsum("tli, tlj, kij -> tkl", n, n, e_plus) / 2  # (t, loc, link)
    g_cross = np.einsum("tli, tlj, kij -> tkl", n, n, e_cross) / 2  # (t, loc, link)

    # Broadcast ltts over frequencies and locations
    ltts = ltts[:, None, None, :]  # (t, f, loc, link)

    # Compute M factor, eq. 2.14
    one_plus_kl = 1 + np.einsum("ki, tli -> tkl", k, n)  #  (t, loc, link)
    m_exp_factor = np.exp(
        np.pi * 1j * ltts * np.einsum("f, tkl -> tfkl", f, one_plus_kl)
    )  # (t, f, loc, link)
    m_sinc = np.sinc(
        ltts * np.einsum("f, tkl -> tfkl", f, one_plus_kl)
    )  # (t, f, loc, link)
    m_factor = m_exp_factor * m_sinc  # (t, f, loc, link)

    # Compute Xi, eq. 2.13
    xi_exp_factor = np.exp(
        -2j * np.pi * np.einsum("f, ki, tli -> tfkl", f, k, l_ij)
    )  # (t, f, loc, link)
    xi_plus = np.einsum(
        "tfkl, tfkl, tkl -> tfkl", xi_exp_factor, m_factor, g_plus
    )  # (t, f, loc, link)
    xi_cross = np.einsum(
        "tfkl, tfkl, tkl -> tfkl", xi_exp_factor, m_factor, g_cross
    )  # (t, f, loc, link)
    xi = np.stack([xi_plus, xi_cross], axis=-1)  # (t, f, loc, link, pol)

    # Broadcast f over locations and links
    f_broadcasted = f[:, None, None]  # (f, loc, link)

    # Add in final exponential factor in eq. 2.16
    # and f^2 pre-factor from eq. 2.15
    final_exp = (
        (2j * np.pi * f_broadcasted * ltts)  # (t, f, loc, link)
        * np.exp(-2j * np.pi * f_broadcasted * ltts)  # (t, f, loc, link)
        * np.exp(
            -2j
            * np.pi
            * np.einsum("f, kx, tlx -> tfkl", f, k, positions[:, index_receiver])
        )  # (t, f, loc, link)
    )  # (t, f, loc, link)

    return final_exp[..., None] * xi  # (t, f, loc, link, pol)


def sky_average(
    cov_func: Callable[[np.ndarray, np.ndarray], np.ndarray],
    *,
    nside: int = 12,
    axis: int = 2,
) -> np.ndarray:
    r"""Compute sky average of a covariance matrix.

    The ``cov_func`` argument can be any function that takes :math:`\beta` and
    :math:`\lambda` (sky localization angles, as latitude and longitude) as
    keyword arguments (with keywords `beta` and `lamb`) and returns an
    covariance matrix where ``axis`` corresponds to the sky pixels.

    This function calls ``cov_func`` for an array of pixels of a Healpix grid,
    and then the resulting covariance matrices will be averaged over the sky.

    .. rubric:: Usage

    .. code:: python

        from segwo.cov import project_covariance

        strain_cov = np.eye(2) / 2  # assumed unpolarized strain, shape (pol, pol)
        link2observables = np.array(...)  # shape (f, obs, link)

        def observable_cov(beta, lamb):
            strain2link = compute_strain2link(f, beta, lamb, ltts, positions)
            return project_covariance(strain_cov, [strain2link, link2observables])

        averaged_observable_cov = compute_sky_average(observable_cov)  # (t, f, obs)

    .. warning::

        Make sure to compute the covariance matrices before averaging them over
        the sky, as pixels should be incoherently summed.

    Args:
        cov_func: Callable that returns covariance matrices, taking ``beta`` and
            ``lamb`` as keyword arguments and returning an array.
        nside: Healpix nside parameter (sets sky resolution).
        axis: Axis of the covariance matrix that corresponds to the sky pixels.

    Returns:
        Sky-averaged covariance matrix, of shape matching the output of
        ``cov_func`` but with the ``axis`` dimension removed (averaged over).

        The resulting units are those of the ``cov_func`` output.
    """
    npix = hp.nside2npix(nside)
    thetas, phis = hp.pix2ang(nside, np.arange(npix))
    # Conversion from colatitude to latitude
    betas, lambs = np.pi / 2 - thetas, phis

    response = cov_func(betas, lambs)
    if response.shape[axis] != npix:
        raise ValueError("Covariance matrix does not have the expected shape.")

    return np.sum(response, axis=axis) / npix


def compute_isotropic_signal_cov(
    f: np.ndarray,
    ltts: np.ndarray,
    positions: np.ndarray,
    *,
    nside: int = 12,
    mixings: None | np.ndarray | list[np.ndarray] = None,
    method: str = "baghi+23",
    strain_cov: np.ndarray | None = None,
) -> np.ndarray:
    r"""Compute signal covariance from an isotropic background [1/Hz].

    This function returns the covariance matrix of a signal for a given set of
    observables computed from an isotropic stochastic gravitational-wave
    background in units of fractional frequency deviation (ie., :math:`\Delta
    \nu / \nu`) spectral power density.

    It is obtained by projecting the strain covariance matrix ``strain_cov``
    (expressed in the plus-cross polarization basis) onto the chosen observables
    (defined by the mixing matrices ``mixings`` going from single linkes to the
    chosen set of observables), using the sky position-dependent link responses;
    then, these covariance matrices are averaged over the sky using the Healpix
    pixelization of the sky.

    .. rubric:: Usage

    .. code:: python

        from segwo.response import compute_isotropic_signal_cov

        f = np.array(...)  # frequency grid [Hz]
        ltts = np.array(...)  # light travel times [s]
        positions = np.array(...)  # test-mass positions [m]

        mixing1 = np.array(...)  # mixing matrix 1
        mixing2 = np.array(...)  # mixing matrix 2

        response = compute_isotropic_signal_cov(
            f,
            ltts,
            positions,
            mixings=[mixing1, mixing2],
        )

    .. admonition:: Detector response

        By default, this function uses a normalized and unpolarized (i.e.,
        diagonal) strain covariance matrix. As a consequence, the covariance
        matrix returned by this function can be interpreted as the incoherent
        sum of the detector response for both polarizations. In this case, the
        units are fractional frequency deviation power spectral density (the
        signal) divided by the strain power spectral density (here equal to 1).

        The covariance matrix can be directly used as input to the sensitivity
        computation :func:`compute_sensitivity_from_covariances`.

    .. seealso:: :func:`segwo.response.strain2link`,
        :func:`segwo.response.sky_average`,
        :func:`segwo.sensitivity.compute_sensitivity_from_covariances`

    Parameters
    ----------
    f : Array of shape (Nf,)
        Frequency grid [Hz].
    ltts : Array of shape (Nt, link=6)
        Light travel times between the test masses [s].
    positions : Array of shape (Nt, sc=3, xyz=3)
        Test-mass positions [m].
    nside : int
        Healpix nside parameter (sets sky resolution).
    mixings : None, array, or list of arrays
        Mixing matrices transforming from single links to the desired set of
        observables.

        If None, the function will return the sky-averaged signal covariance at
        for single links.

        If a single array, it should be broadcastable to ``(Nt, Nf, Nobs,
        link=6)``, where ``Nobs`` is the number of observables. If a list of
        arrays, they should be of compatible shapes, such that the overall
        mixing matrix has the shape mentioned above (see
        :func:`segwo.cov.compose_mixings`).
    method : str
        Computation method, either "baghi+23" or "hartwig+23".

        See :func:`compute_strain2link` for more details.
    strain_cov : Array of shape (Npol=2, Npol=2)
        Strain covariance matrix.

        If None, the function will use a normalized and unpolarized strain
        (diagonal) covariance matrix.

    Returns
    -------
    np.ndarray of shape (Nt, Nf, Nobs, Nobs)
        Sky-averaged signal covariance matrix.

        Units are fractional frequency deviation power spectral density [1/Hz].
    """
    # By default, use a normalized, unpolarized strain covariance matrix
    if strain_cov is None:
        strain_cov = np.eye(2) / 2  # (pol, pol)

    # Make sure we have a list of mixings
    if mixings is None:
        mixings = []
    if isinstance(mixings, np.ndarray):
        mixings = [mixings]
    # Insert an extra dimansion in the mixing matrices for sky locations
    mixings = [mixing[..., None, :, :] for mixing in mixings]

    # Compute the response covariance for a specific sky location
    def obs_cov(beta, lamb) -> np.ndarray:
        strain2link = compute_strain2link(
            f, beta, lamb, ltts, positions, method=method
        )  # (t, f, loc, obs, obs)
        strain2obs = [strain2link] + mixings  # (t, f, loc, obs, obs)
        return project_covariance(strain_cov, strain2obs)  # (t, f, loc, obs, obs)

    # Average over the sky
    averaged_obs_cov = sky_average(obs_cov, nside=nside, axis=2)  # (t, f, obs, obs)
    return averaged_obs_cov  # (t, f, obs, obs)



"""
Symbolic expressions
====================

This module offers functions to manipulate symbolic expressions of
covariance and mixing matrices. It mirrors the functionality of the
:mod:`segwo.cov` module, but operates on symbolic expressions instead of
numerical arrays. This allows to compute analytic expressions for the noise
transfer functions under certain assumptions.

Symbols
-------

.. autodata:: f

.. autodata:: d

.. autofunction:: delay_exp

Building covariance matrices
----------------------------

.. autofunction:: construct_covariance_from_psds

.. autofunction:: concatenate_covariances

Building mixing matrices
------------------------

.. autofunction:: construct_mixing_from_pytdi

.. autofunction:: compose_mixings

.. autofunction:: concatenate_mixings

Projecting measurements
-----------------------

.. autofunction:: project_covariance

"""

import pytdi
import sympy as sp

f = sp.symbols("f", real=True, positive=True)
"""Frequency symbol [Hz]."""

d = {
    link: sp.symbols(f"d_{{{link}}}", real=True, positive=True)
    for link in [12, 23, 31, 13, 32, 21]
}
"""Dictionary of light travel times (delays) for each link [s]."""


def delay_exp(links: int | list[int]) -> sp.Expr:
    r"""Compute the delay as a complex exponential.

    Assuming that delays are constnts, they can be written in the frequency
    domain as a complex exponential,

    .. math::

        e^{-i 2 \pi f d_{ij}}.

    This function makes use of :data:`f` and :data:`d`.

    Parameters
    ----------
    links : int or list of int
        The links to compute the delay for. If a single link is given, it is
        assumed to be a single link. Otherwise, it is assumed to be a list of
        links and the delays are simply added together.

    Returns
    -------
    sympy.Expr
        The delay as a complex exponential.
    """
    if isinstance(links, int):
        links = [links]

    return sp.prod(sp.exp(-2j * sp.pi * f * d[link]) for link in links)


def construct_covariance_from_psds(psds: list[sp.Expr]) -> sp.Matrix:
    """Construct covariance matrix from a list of power spectral densities.

    We assume that the covariance matrix is diagonal (i.e. no correlations).

    Parameters
    ----------
    psds : list of N sympy.Expr
        List of power spectral densities. The order of the PSDs in this list
        determines the order of the rows and columns in the covariance matrix.

    Returns
    -------
    sympy.Matrix
        Covariance matrix, of shape ``N x N``.
    """
    return sp.diag(*psds)


def concatenate_covariances(covariances: list[sp.Matrix]) -> sp.Matrix:
    r"""Concatenate a sequence of covariance matrices into a single matrix.

    See :func:`segwo.cov.concatenate_covariances` for a description of the
    function.

    Parameters
    ----------
    covariances : list of N sympy.Matrix
        Covariance matrices of shapes ``N[i] x N[i]``, where ``N[i]`` is the
        number of observables in the i-th covariance matrix.

    Returns
    -------
    sympy.Matrix
        Concatenated covariance matrix, of shape ``N x N``, where ``N`` is the
        total number of observables across all covariance matrices.

        Units of each covariance matrix remain the same.
    """
    # Check we have at least one covariance matrix
    if not covariances:
        return sp.Matrix([])
    assert len(covariances) > 0

    block_matrix = sp.BlockDiagMatrix(*covariances)
    matrix = sp.Matrix(block_matrix)

    return matrix


def _pytdi_delay_exp(delays: list[str]) -> sp.Expr:
    """Transform a list of delay strings from PyTDI to sympy expressions.

    PyTDI uses a different syntax for delays, so we need to convert them to
    sympy expressions. This function is used internally in the
    :func:`construct_mixing_from_pytdi` function.

    PyTDI uses a list of strings, such as "A_12" or "D_31" to represent delays
    and advancements. Because we assume that delays are constant, we can
    simply replace these strings with the corresponding sympy symbols and
    advancements are replaced with the corresponding negative delays.

    Parameters
    ----------
    delays : list of str
        List of strings representing delays and advancements. The strings can
        be in the form "A_ij" or "D_ij", where "ij" is the link number.

    Returns
    -------
    sympy.Expr
        The delay as a complex exponential.
    """
    # Convert the list of strings to a list of sympy symbols
    delays = [
        d[int(link[2:])] if link.startswith("D_") else -d[int(f"{link[-1]}{link[-2]}")]
        for link in delays
    ]

    # Compute the product of the delays
    return sp.prod(sp.exp(-2j * sp.pi * f * delay) for delay in delays)


def construct_mixing_from_pytdi(
    tdi_combinations: list[pytdi.TDICombination],
    measurements: list[str],
) -> sp.Matrix:
    """Construct the mixing matrix from the TDI combinations.

    See :func:`segwo.cov.construct_mixing_from_pytdi` for a description of the
    function.

    Parameters
    ----------
    tdi_combinations : list of N pytdi.TDICombination
        List of TDI combinations. The order of the combinations in this list
        determines the order of the rows in the mixing matrix.
    measurements : list of M str
        List of measurement names. The order of the measurements in this list
        determines the order of the columns in the mixing matrix.

    Returns
    -------
    sympy.Matrix
        Mixing matrix, of shape ``N x M``.
    """
    # Start with an empty matrix
    mixing = sp.zeros(len(tdi_combinations), len(measurements))

    # For each combination, compute the corresponding row of the mixing matrix
    for i, combination in enumerate(tdi_combinations):
        # Fill element i,j of mixing matrix if measurement is in combination
        for j, measurement in enumerate(measurements):
            if measurement in combination.components:
                # Couplings are a list of tuples (coeff, list of delays)
                couplings = combination.components[measurement]
                mixing[i, j] = sum(
                    coeff * _pytdi_delay_exp(delays) for coeff, delays in couplings
                )

    return mixing


def compose_mixings(mixings: list[sp.Matrix]) -> sp.Matrix:
    r"""Compose a sequence of mixing matrices into a single mixing matrix.

    See :func:`segwo.cov.compose_mixings` for a description of the function.

    Parameters
    ----------
    mixings : list of N sympy.Matrix
        Mixing matrices of compatible successive shapes. The i-th mixing matrice
        should have shape ``M[i+1] x M[i])``.

    Returns
    -------
    sympy.Matrix
        Composed mixing matrix, of shape ``M[N] x M[0]``.

        Units are the product of the units of the input mixing matrices.

    Raises
    ------
    ValueError
        If the mixing matrices are not compatible, or if there are no mixing
        matrices.
    """
    # Check we have at least one mixing matrix
    if not mixings:
        raise ValueError("mixings must be a non-empty list.")

    return sp.prod(mixings[::-1])


def concatenate_mixings(mixings: list[sp.Matrix]) -> sp.Matrix:
    r"""Concatenate a sequence of mixing matrices into a single mixing matrix.

    See :func:`segwo.cov.concatenate_mixings` for a description of the function.

    Parameters
    ----------
    mixings : list of N sympy.Matrix
        Mixing matrices of shapes ``N[i] x N[i]``, where ``N[i]`` is the number
        of observables in the i-th mixing matrix.

    Returns
    -------
    sympy.Matrix
        Concatenated mixing matrix, of shape ``N x N``, where ``N`` is the
        total number of observables across all mixing matrices.

        Units of each mixing matrix remain the same.
    """
    # Check we have at least one mixing matrix
    if not mixings:
        return sp.Matrix([])
    assert len(mixings) > 0

    block_matrix = sp.BlockDiagMatrix(*mixings)
    matrix = sp.Matrix(block_matrix)

    return matrix


def project_covariance(
    cov: sp.Matrix, mixing: sp.Matrix | list[sp.Matrix]
) -> sp.Matrix:
    """Transform covariance to a new set of observables using mixing matrices.

    See :func:`segwo.cov.project_covariance` for a description of the function.

    Parameters
    ----------
    cov : sympy.Matrix
        Covariance matrix of shape ``N x N``.
    mixing : sympy.Matrix or list of sympy.Matrix
        Mixing matrix or list of mixing matrices. If a list is provided, the
        mixing matrices are composed before applying the transformation.

        The shape of the composed matrix must be compatible with the covariance
        matrix, i.e., it should be of shape ``M x N``, where ``M`` is the number
        of observables in the new set of observables.

    Returns
    -------
    sympy.Matrix
        Transformed covariance matrix of shape ``M x M``, where ``M`` is the
        number of observables in the new set of observables.

        Units are the product of the units of the input covariance and mixing
        matrices.
    """
    # Check we have at least one mixing matrix
    if not isinstance(mixing, sp.Matrix):
        mixing = compose_mixings(mixing)

    return mixing * cov * mixing.H



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""SEGWO package."""

import importlib.metadata

from . import cov, response, sensitivity, sym

__copyright__ = "2025, Olaf Hartwig and Jean-Baptiste Bayle"
__author__ = "Olaf Hartwig and Jean-Baptiste Bayle"
__version__ = importlib.metadata.version("segwo")



r"""
Covariance matrices
===================

SEGWO provides a set of convenient functions to define frequency-domain
covariance matrices. Covariance matrices capture statistical properties of (at
least locally) stationary, potentially correlated processes.

Mixing matrices define the linear transformation between a set of observables
to another set of observables. They are used to project measurements or
covariance matrices from one set of observables to another.

Building covariance matrices
----------------------------

In our case, these processes are often noises, but they can also be
gravitational-wave signals. A covariance matrix :math:`\mathbf{C}` then
describes how theses noises (or signals) appear in a given set of observables
(e.g., a set of measurements). For a set of :math:`N` observables :math:`y_i`,
the covariance matrix is a :math:`N \times N` matrix.

Entries :math:`\mathbf{C}_{ii}` on the diagonal of a covariance matrix
correspond to the real-valued power spectral density (PSD) of the process in
each observable :math:`i`. Off-diagonal entries :math:`\mathbf{C}_{ij}`
correspond to the complex-valued cross-spectral density (CSD) between
observables :math:`i` and :math:`j`. Note that covariance matrices are always
Hermitian and positive-definite, such that :math:`\mathbf{C}_{ij} =
\mathbf{C}_{ji}^*`.

As mentioned above, SEGWO works in the frequency domain, under the assumption of
local stationarity. That means that covariance matrices are time _and_ frequency
dependent. Users are expected to define a time grid (with :math:`N_t` time
steps) and a frequency grid (with :math:`N_f` frequency steps).

The resulting covariance matrix is then a 4D Numpy array of shape :math:`(N_t,
N_f, N, N)`, where the first two dimensions are used for time and frequency, and
the last two dimensions are used for the observables.

One way to build a covariance matrix for a set of uncorrelated observables is
to provide a list of PSD values for each observable. The covariance matrix can
be built by calling :func:`construct_covariance_from_psds`, and will be
a diagonal matrix with the PSD values on the diagonal.

.. code:: python

    import numpy as np
    from segwo.cov import construct_covariance_from_psds

    # Frequency grid
    f = np.logspace(...)  # [Hz]

    # Define a list of PSDs for each observable
    psd1 = 1e-20 * (1 + (2e-3 / f) ** 4)  # Example PSD
    psd2 = 1e-15 * 2 * np.pi * f * (1 + (f / 8e-3) ** 4)  # Example PSD
    psd3 = 1e-20 * (1 + (2e-3 / f) ** 4)  # Example PSD

    # Build the covariance matrix
    noise_cov = construct_covariance_from_psds([psd1, psd2, psd3])


You can also concatenate existing covariance matrices into a single covariance
matrix. This is useful when you have a set of covariance matrices for different,
uncorrelated observables, and you want to combine them as a block-diagonal
matrix. You can use the :func:`concatenate_covariances` function to do this.


.. autofunction:: construct_covariance_from_psds

.. autofunction:: construct_covariance_from_measurements

.. autofunction:: concatenate_covariances

Building mixing matrices
------------------------

Mixing matrices define the linear transformation between a set of :math:`M`
observables to another set of :math:`N` observables. They are defined in the
frequency domain as :math:`N \times M` complex-valued matrices under the
assumption of local stationarity.

As a consequence, they are also time- and frequency-dependent. The mixing
matrices are defined for each time step and frequency step, and the resulting
mixing matrix is a 4D Numpy array of shape :math:`(N_t, N_f, N, M)`, where the
first two dimensions are used for time and frequency, and the last two
dimensions are used for the observables.

One way to build a mixing matrix is to provide a list of TDI combinations
defining the transfer functions for each observable. The TDI combinations are
defined as a dictionary of the form ``{measurement: [(scaling_factor, [delays]),
...]}``, where the measurement is the input observable, the scaling factor is a
complex number, and the delays are lists of delays/advancements to be applied to
the measurement. The mixing matrix can be built by calling
:func:`construct_mixing_from_pytdi`.

.. code:: python

    from segwo.cov import construct_mixing_from_pytdi
    import numpy as np
    import pytdi

    # Frequency grid
    f = np.logspace(...)  # [Hz]

    # Define a list of TDI combinations for each observable
    tdi_combinations = [
        pytdi.TDICombination(
            "eta_31", [(-1, ["A_21", "D_13"]), (1, ["D_21", "D_13"])],
            "eta_13", [(-1, ["A_21"]), (1, ["D_21"])],
        ),
        pytdi.TDICombination(
            "eta_13", [(-1, ["A_21"]), (1, ["D_21"])],
            "eta_12", [(1, ["A_21"])],
            "eta_21", [(1, [])],
        ),
        pytdi.TDICombination(
            "eta_21", [(1, [])],
            "eta_12", [(1, ["A_21"])],
        ),
    ]

    # Define the list of ordered original measurements
    measurements = ["eta_31", "eta_13", "eta_21", "eta_12"]

    # Define a list of light travel times for each link at each time point
    ltts = np.array(...) #  shape (t, link)

    # Build the mixing matrix
    mixing = construct_mixing_from_pytdi(
        f, measurements, tdi_combinations, ltts
    )

You can also compose a sequence of mixing matrices into a single mixing matrix,
to represent a sequence of transformations, with :func:`compose_mixings`. This
is useful when you have a set of mixing matrices for different observables and
you want to represent them as a single transformation to go from the first set
of observables to the last set.

.. autofunction:: construct_mixing_from_pytdi

.. autofunction:: compose_mixings

.. autofunction:: concatenate_mixings

Projecting measurements
-----------------------

.. autofunction:: project_measurements

.. autofunction:: project_covariance

"""

from typing import TypeAlias

import numpy as np
import pytdi
from numpy.typing import DTypeLike


def compose_mixings(mixings: list[np.ndarray]) -> np.ndarray:
    r"""Compose a sequence of mixing matrices into a single mixing matrix.

    This function takes a sequence of N mixing matrices :math:`\mathbf{M}^i`
    and composes them into a single mixing matrix
    :math:`\mathbf{M}`, such that the first mixing matrix in the sequence is
    applied first. Ie., we construct the matrix product

    .. math::

        \mathbf{M} = \mathbf{M}^{N-1}
        \mathbf{M}^{N-2} \ldots \mathbf{M}^0.

    Args:
        mixings: List of N mixing matrices, of compatible successive shapes. The
            i-th mixing matrice should have shape ``(..., M[i+1], M[i])``.

    Returns:
        Composed mixing matrix, of shape ``(..., M[N], M[0])``.

        Units are the product of the units of the input mixing matrices.
    """
    # Check we have at least one mixing matrix
    if not mixings:
        raise ValueError("mixings must be a non-empty list.")

    # If we have only one mixing matrix, return it
    if len(mixings) == 1:
        return mixings[0]

    # Recursively compose the mixing matrices
    mixing_new = mixings[-1]  # (..., M[N], M[N-1))
    mixing_recursed = compose_mixings(mixings[:-1])  # (..., M[N-1], M[0])

    return np.einsum(
        "...ij, ...jk -> ...ik", mixing_new, mixing_recursed
    )  # (..., M[N], M[0])


def concatenate_mixings(
    mixings: list[np.ndarray], *, dtype: DTypeLike | None = None
) -> np.ndarray:
    r"""Concatenate a sequence of mixing matrices into a single mixing matrix.

    This function takes a sequence of N mixing matrices :math:`\mathbf{M}^i` and
    concatenates them along the last 2 axes into a single mixing matrix
    :math:`\mathbf{M}`. We assume the observables across mixing matrices are
    uncorrelated; therefore, we effectively form the block-diagonal mixing
    matrix

    .. math::

        \mathbf{M} = \begin{pmatrix}
            \mathbf{M}^{0} & & & \\
            & \mathbf{M}^{1} & & \\
            & & \ddots & \\
            & & & \mathbf{M}^{N-1}
        \end{pmatrix}.

    .. rubric:: Usage

    .. code:: python

        mixing1 = np.array(...)
        mixing2 = np.array(...)

        concatenated_mixing = concatenate_mixings([mixing1, mixing2])

    Args:
        mixings:
            List of N mixing matrices, of shapes ``(..., N[i], N[i])``, where
            ``N[i]`` is the number of observables in the i-th mixing matrix. The
            first dimensions are traditionally used for time and frequency, and
            should be broadcastable across the mixing matrices.
        dtype:
            Data type of the output mixing matrix. If not provided, the data
            type of the first mixing matrix is used.

    Returns:
        Concatenated mixing matrix, of shape ``(..., N, N)``, where ``N`` is the
        total number of observables across all mixing matrices. The first
        dimensions are traditionally used for time and frequency, and are kept
        in the output.

        Units of each mixing matrix remain the same.
    """
    # Check we have at least one mixing matrix
    if not mixings:
        return np.empty((0, 0), dtype=float)
    assert len(mixings) > 0

    # Use the data type of the first mixing matrix if not provided
    if dtype is None:
        dtype = mixings[0].dtype

    # Form the block-diagonal mixing matrix
    total_size = sum(mixing.shape[-1] for mixing in mixings)
    first_dimensions = [mixing.shape[:-2] for mixing in mixings]
    broadcasted_first_dimensions = np.broadcast_shapes(*first_dimensions)
    result = np.zeros(
        (*broadcasted_first_dimensions, total_size, total_size),
        dtype=dtype,
    )

    # Fill the block-diagonal mixing matrix
    current_index = 0
    for mixing in mixings:
        size = mixing.shape[-1]
        end_index = current_index + size
        result[..., current_index:end_index, current_index:end_index] = mixing
        current_index = end_index

    # Check that we filled the entire matrix
    if current_index != total_size:
        raise ValueError(
            "Not all mixing matrices were filled in the block-diagonal matrix."
        )

    return result  # (..., N, N)


def construct_mixing_from_pytdi(
    f: np.ndarray,
    measurements: list[str],
    tdi_combinations: (
        list[pytdi.TDICombination] | list[list[pytdi.TDICombination]]
    ),  # (tdi,) or (t, tdi)
    ltts: np.ndarray,
) -> np.ndarray:
    r"""Construct the mixing matrix from a list of TDI combinations.

    This function calls :func:`_construct_mixing_vector_from_pytdi_components`,
    but handles multiple TDI combinations at each time point. It also extracts
    the TDI components from the TDI combinations.

    Typical use case would be to pass a set of combinations, for example, X, Y,
    Z, at each time point.

    Args:
        f:
            Frequency array [Hz].
        measurements:
            Ordered list of measurements.
        tdi_combinations:
            List of TDI combinations for one time point, or list of lists of TDI
            combinations.
        ltts:
            Light travel times for each link at each time point, of shape
            ``(t, link)``.

    Returns:
        The constructed mixing matrix with shape ``(t, f, tdi, measurements)``.
    """
    # Check that `tdi_combinations` is an homogeneous list
    if not all(isinstance(tc, list) for tc in tdi_combinations) and not all(
        isinstance(tc, pytdi.TDICombination) for tc in tdi_combinations
    ):
        raise ValueError(
            "All elements of tdi_combinations must be lists "
            "or TDICombination objects."
        )
    # If this is a list of TDICombination objects, we need to convert it
    # to a list of lists
    if isinstance(tdi_combinations[0], pytdi.TDICombination):
        tdi_combinations = [tdi_combinations] * ltts.shape[0]
    # Else, check that every element is a list of TDICombination objects
    else:
        for tdi_combination in tdi_combinations:
            if not all(isinstance(tc, pytdi.TDICombination) for tc in tdi_combination):
                raise ValueError(
                    "All elements of tdi_combinations must be lists "
                    "of TDICombination objects."
                )

    # Extract components from TDI combinations
    components = [
        [combination.components for combination in timestep]
        for timestep in tdi_combinations
    ]

    # Stack the results to get a final shape (t, f, tdi, measurements)
    results = []
    for i in range(len(components[0])):
        # Extract the i-th slice of shape (t,)
        slice_input = [components[j][i] for j in range(len(components))]

        result = _construct_mixing_vector_from_pytdi_components(
            f, measurements, slice_input, ltts
        )

        # Store the result
        results.append(result)

    # Stack along the tdi axis
    return np.stack(results, axis=2)


# pylint: disable=invalid-name
PyTDIComponent: TypeAlias = dict[str, list[tuple[float, list[str]]]]
# pylint: enable=invalid-name


def _construct_mixing_vector_from_pytdi_components(
    f: np.ndarray,
    measurements: list[str],
    tdi_components: PyTDIComponent | list[PyTDIComponent],
    ltts: np.ndarray,
) -> np.ndarray:
    r"""Construct mixing vector for a single time-dependent TDI observable.

    PyTDI defines TDI combinations as a dictionary of the form

    .. code-block:: python

        {
            'eta_31': [(-1, ['A_21', 'D_13']), (1, ['D_21', 'D_13'])],
            'eta_13': [(-1, ['A_21']), (1, ['D_21'])],
            'eta_21': [(1, [])],
            'eta_12': [(1, ['A_21'])],
        }

    where the item labels denote the measurement to be delayed and the values
    are lists of touples of scaling factors and accompanying delays/advancements
    to be applied to the measurement.

    Assuming delays to be constant, we can convert the list of delays into
    Fourier domain exponential factors.

    Advancement operators satisfy ``D_ij A_ji = 1``, such that the value of each
    advancement is computed as ``a_ji = -d_ij``.

    The corresponding Fourier domain factors are then given as

    .. math::

        \mathrm{FT}[\mathbf{D}_{ij} x] &= e^{- 2 \pi f d_{ij}(t)} \mathrm{FT}[x],

        \mathrm{FT}[\mathbf{A}_{ij} x] &= e^{2 \pi f d_{ji}(t)} \mathrm{FT}[x].

    Parameters:
        f:
            Frequency array [Hz].
        measurements:
            Ordered list of measurements.
        tdi_components:
            Dictionary containing delays for measurement
            (e.g., 'eta_21': [(-1, ['D_12']), (1, ['D_13', 'D_31', 'D_12'])]),
            or list of such dictionaries for each time point in ``ltts``.
        ltts:
            Light travel times for each link at each time point, of shape
            ``(t, link)``.

    Returns:
        The constructed mixing vector with shape ``(t, f, measurements)``.
    """
    if ltts.ndim != 2:
        raise ValueError("ltts must be a 2D array of shape (t, link).")
    if ltts.shape[1] != 6:
        raise ValueError("ltts must have 6 columns, one for each link.")

    delays = ltts  # (t, link)
    links = [12, 23, 31, 13, 32, 21]  # (link,)
    # Index order is 12, 23, 31, 13, 32, 21,
    # ie., advancements simply flip the array
    advancements = np.flip(delays, axis=-1)  # (t, link)

    # Convert tdi_components to a list of dictionaries
    if not isinstance(tdi_components, list):
        tdi_components = [tdi_components] * delays.shape[0]
    if len(tdi_components) != delays.shape[0]:
        raise ValueError("Length of tdi_components must match length of delays.")

    def compute_exponentials(delay_list: list[str], t_index: int) -> np.ndarray:
        """Compute frequency domain factor corresponding to list of delays.

        Args:
            delay_list: List of delays and advancements for a specific
                measurement at a specific time point (index ``t_index``).
            t_index: Index of the time point.

        Returns:
            Frequency domain factor as a complex array of shape
            ``(len(f))``.
        """
        result = np.ones(len(f), dtype=complex)  # (f,)
        for item in delay_list:
            if item[0] == "A":
                result *= np.exp(
                    1j
                    * 2
                    * np.pi
                    * f  # (f,)
                    * advancements[t_index, links.index(int(item[2:]))]
                )
            elif item[0] == "D":
                result *= np.exp(
                    -1j
                    * 2
                    * np.pi
                    * f  # (f,)
                    * delays[t_index, links.index(int(item[2:]))]
                )
            else:
                raise ValueError(f"Invalid delay list entry: {item}")

        return result  # (f,)

    def build_factor_from_component(measurement: str, t_index: int) -> np.ndarray:
        """Computes frequency domain factor for given measurement.

        Args:
            measurement: Measurement label.
            t_index: Index of the time point.

        Returns:
            Frequency domain factor as a complex array of shape
            ``(len(f))``.
        """
        result = np.zeros(len(f), dtype=complex)  # (f,)
        for item in tdi_components[t_index][measurement]:
            result += item[0] * compute_exponentials(item[1], t_index)  # (f,)

        return result  # (f,)

    final = np.zeros((delays.shape[0], len(f), len(measurements)), dtype=complex)

    for t_index, t_tdi_components in enumerate(tdi_components):
        for meas_index, measurement in enumerate(measurements):
            if measurement in t_tdi_components:
                final[t_index, :, meas_index] = build_factor_from_component(
                    measurement, t_index
                )  # (f,)

    return final  # (t, f, measurements)


def project_measurements(x: np.ndarray, mixing: np.ndarray) -> np.ndarray:
    r"""Transform measurements to a new set of observables using a mixing matrix.

    This is a convenience function to apply a mixing matrix to a set of
    observables :math:`\mathbf{x}_j`, with shape ``(..., M)``. First axis are
    traditionally used for time and frequency, and are kept in the output.

    The mixing matrix :math:`\mathbf{M}` is always applied to the last
    dimension of the input array. The mixing matrix must be of shape ``(..., N,
    M)``.

    .. math::

        \mathbf{y}_{i} = \mathbf{M}_{ij} \mathbf{x}_{j}.

    Args:
        x: Array of observables, of shape ``(..., M)``.
        mixing: Mixing matrix, of shape ``(..., N, M)``.

    Returns:
        Projected observables, of shape ``(..., N)``.

        Units are the product of the units of ``x`` and ``mixing``.
    """
    # Number of dimensions for x and mixing
    if x.ndim < 1:
        raise ValueError("x must have at least 1 dimension.")
    if mixing.ndim < 2:
        raise ValueError("mixing must have at least 2 dimensions.")

    return np.einsum("...ij, ...j -> ...i", mixing, x)  # (..., meas)


def construct_covariance_from_measurements(x: np.ndarray) -> np.ndarray:
    r"""Compute covariance matrices from a vector of measurements.

    The input vector is an array of measurements :math:`\mathbf{x}_{i}`, with
    shape ``(..., N)``. The last dimension is reserved for the measurements, and
    additional first dimensions (traditionally for time and frequency) are kept
    in the output.

    The covariance matrix :math:`\mathbf{C}^{\mathbf{x}}_{ij}` is computed by
    taking the outer product of the measurement vector with its complex
    conjugate on the last dimension,

    .. math::

        \mathbf{C}^{\mathbf{x}}_{ij} = \mathbf{x}_{i}
        \mathbf{x}_{j}^{*}.

    .. warning::

        Adding covariance matrices is not equivalent to computing the covariance
        matrix of the summed measurement vectors. Only add covariance matrices
        for quantities that are statistically independent.

    Args:
        x: Array of measurements, of shape ``(..., N)``.

    Returns:
        Covariance matrix, of shape ``(..., N, N)``, with ``N`` the last
        dimension of ``x``. Additional dimensions are kept.

        Units are those of ``x``, squared.
    """
    # Check that x has at least 2 dimensions
    if x.ndim < 1:
        raise ValueError(f"x must have at least 1 dimensions, got {x.ndim}.")

    return np.einsum("...i, ...j -> ...ij", x, np.conj(x))  # (..., meas, meas)


def construct_covariance_from_psds(psds: list[np.ndarray]) -> np.ndarray:
    r"""Construct covariance matrix from a list of power spectral densities.

    This function takes a list of frequency-dependent PSDs :math:`S_i(f)` for
    each observable of interest :math:`i` and constructs the associated mixing
    matrix :math:`\mathbf{C}(f)`. The observables are assumed to be
    uncorrelated, such at the covariance matrix is diagonal.

    .. math::

        \mathbf{C}^{\mathbf{x}}_{ij}(f) = \mathbf{S}_i(f) \delta_{ij}.

    .. rubric:: Usage

    .. code:: python

        f = np.array(...)  # Frequency array [Hz]
        psd1 = 1e-20 * (1 + (2e-3 / f) ** 4) # Example PSD
        psd2 = 1e-15 * 2 * np.pi * f * (1 + (f / 8e-3) ** 4) # Example PSD

        noise_cov = construct_covariance_from_psds([psd1, psd2])

    Args:
        psds:
            List of N PSDs, each as an array of shape ``(..., f)``. The first
            dimensions are traditionally used for time and frequency, and are
            kept in the output. The last dimension is reserved for the

    Returns:
        Mixing matrix, of shape ``(t, f, obs, obs)``.

        Units are the product of the units of ``psds``.
    """
    # Turn PSD arrays into single-observable covariance matrices
    covs = [psd[..., None, None] for psd in psds]
    return concatenate_covariances(covs)


def concatenate_covariances(
    covs: list[np.ndarray], *, dtype: DTypeLike | None = None
) -> np.ndarray:
    r"""Concatenate covariance matrices into a single covariance matrix.

    This function takes a sequence of N covariance matrices :math:`\mathbf{C}^i` and
    concatenates them along the last 2 axes into a single covariance matrix
    :math:`\mathbf{C}`. We assume the observables across covariance matrices are
    uncorrelated; therefore, we effectively form the block-diagonal covariance
    matrix

    .. math::

        \mathbf{C} = \begin{pmatrix}
            \mathbf{C}^{0} & & & \\
            & \mathbf{C}^{1} & & \\
            & & \ddots & \\
            & & & \mathbf{C}^{N-1}
        \end{pmatrix}.

    .. rubric:: Usage

    .. code:: python

        cov1 = np.array(...)
        cov2 = np.array(...)

        concatenated_cov = concatenate_covariances([cov1, cov2])

    Args:
        covs:
            List of N covariance matrices, of shapes ``(..., N[i], N[i])``,
            where ``N[i]`` is the number of observables in the i-th covariance
            matrix. The first dimensions are traditionally used for time and
            frequency, and should be broadcastable across the covariance
            matrices.
        dtype:
            Data type of the output covariance matrix. If not provided, the data
            type of the first covariance matrix is used.

    Returns:
        Concatenated covariance matrix, of shape ``(..., N, N)``, where ``N`` is
        the total number of observables across all covariance matrices. The
        first dimensions are traditionally used for time and frequency, and are
        kept in the output.

        Units of each covariance matrix remain the same.
    """
    # Check we have at least one covariance matrix
    if not covs:
        return np.empty((0, 0), dtype=float)
    assert len(covs) > 0

    # Use the data type of the first covariance matrix if not provided
    if dtype is None:
        dtype = covs[0].dtype

    # Form the block-diagonal covariance matrix
    total_size = sum(mixing.shape[-1] for mixing in covs)
    first_dimensions = [mixing.shape[:-2] for mixing in covs]
    broadcasted_first_dimensions = np.broadcast_shapes(*first_dimensions)
    result = np.zeros(
        (*broadcasted_first_dimensions, total_size, total_size),
        dtype=dtype,
    )

    # Fill the block-diagonal covariance matrix
    current_index = 0
    for cov in covs:
        size = cov.shape[-1]
        end_index = current_index + size
        result[..., current_index:end_index, current_index:end_index] = cov
        current_index = end_index

    # Check that we filled the entire matrix
    if current_index != total_size:
        raise ValueError(
            "Not all covariance matrices were filled in the block-diagonal matrix."
        )

    return result  # (..., N, N)


def project_covariance(
    cov: np.ndarray, mixings: np.ndarray | list[np.ndarray]
) -> np.ndarray:
    r"""Transform covariance to a new set of observables using mixing matrices.

    This is a convenience function to apply a mixing matrix to a covariance
    :math:`\mathbf{C}^{\mathbf{x}}_{ij}`, associated with a set of observables
    :math:`\mathbf{x}` and of shape ``(..., M, M)``. The mixing matrix is
    applied to the 2 last dimensions of the input covariance matrix. Additional
    axes can be present (traditionally for time and frequency) and are kept in
    the output.

    The result is a new covariance matrix
    :math:`\mathbf{C}^{\mathbf{y}}_{ij}` for a new set of observables
    :math:`\mathbf{y}`, with shape ``(..., N, N)``,

    .. math::

        \mathbf{C}^{\mathbf{y}}_{ij} = \mathbf{M}_{ik}
        \mathbf{C}^{\mathbf{x}}_{kl} \mathbf{M}_{jl}^{*}.

    The mixing matrix :math:`\mathbf{M}` is defined by

    .. math::

        \mathbf{y}_{i} = \mathbf{M}_{ij} \mathbf{x}_{j}.

    The mixing matrix must be of shape ``(..., N, M)``.

    You can use a list of successive mixing matrices. In this case, this
    function first computes the overall mixing matrix by calling
    :func:`compose_mixings`, and then apply it to the covarance matrix.

    .. seealso:: :func:`compose_mixings`

    Args:
        cov: Input ovariance matrix, of shape ``(..., M, M)``.
        mixings: A single mixing matrix, or a list of N mixing matrices. If a
            single array, it should be of shape ``(..., N, M)``. If a
            list of arrays, they should be of compatible shapes, such that the
            overall mixing matrix has the shape mentioned above.

    Returns:
        Projected covariance matrix, of shape ``(..., N, N)``.

        Units are the product of the units of ``cov`` and square of ``mixing``.
    """
    # If no mixing matrix is given, return the input covariance matrix
    if isinstance(mixings, list) and not mixings:
        return cov

    # Compute overall mixing matrix in case of a list
    if isinstance(mixings, list):
        mixings = compose_mixings(mixings)
    assert isinstance(mixings, np.ndarray)

    # Number of dimensions for cov and mixing
    if cov.ndim < 2:
        raise ValueError("cov must have at least 2 dimensions.")
    if mixings.ndim < 2:
        raise ValueError("mixing must have at least 2 dimensions.")

    return np.einsum(
        "...ij, ...jk, ...lk -> ...il", mixings, cov, np.conj(mixings)
    )  # (..., meas, meas)
